package tech.loedige;

import java.math.BigInteger;
import java.security.KeyPair;
import java.security.KeyPairGenerator;
import java.security.NoSuchAlgorithmException;
import java.security.SecureRandom;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Random;
import java.util.concurrent.Callable;
import java.util.function.Function;

public class Main {

    public static void main(String[] args) throws Exception {
        final int numberOfRuns = 10;
        long start=0;
        long finish = 0;
        List<Long> trialsList = new ArrayList<>();
        long[] trials = new long[numberOfRuns];
        BigInteger[] KeyPair;

        //average the RSA key generation 10 times for 1024, 2048 and 4096 bit using the KeyPairGenerator
        getTimesAndAverage(()->{return generateRSAWithLibrary(1024);},10,"1024 Bit mit KeyPairGenerator");
        getTimesAndAverage(()->{return generateRSAWithLibrary(2048);},10,"2048 Bit mit KeyPairGenerator");
        getTimesAndAverage(()->{return generateRSAWithLibrary(4096);},10,"4096 Bit mit KeyPairGenerator");

        //average the RSA key generation 10 times for 1024, 2048 and 4096 bit using the sef written function
        getTimesAndAverage(()->{return generateRSAWithBigInteger(1024);},10,"1024 Bit mit BigInteger");
        getTimesAndAverage(()->{return generateRSAWithBigInteger(2048);},10,"2048 Bit mit BigInteger");
        getTimesAndAverage(()->{return generateRSAWithBigInteger(4096);},10,"4096 Bit mit BigInteger");

   }

    private static BigInteger[] generateRSAWithLibrary(int bitLength){
        try {
            KeyPairGenerator keyPairGenerator = KeyPairGenerator.getInstance("RSA");
            keyPairGenerator.initialize(bitLength);
            KeyPair keyPair = keyPairGenerator.generateKeyPair();
            return new BigInteger[]{new BigInteger(keyPair.getPrivate().getEncoded()), new BigInteger(keyPair.getPublic().getEncoded())};

        } catch (NoSuchAlgorithmException e) {
            e.printStackTrace();
            return null;
        }
    }

    private static void getTimesAndAverage(Callable<BigInteger[]> function, int numberOfRuns, String functionName) throws Exception {
        long start=0;
        long finish=0;
        List<Long> trialTimesList = new ArrayList<>();
        long[] trialTimes = new long[numberOfRuns];
        BigInteger[] KeyPair;

        for(int i = 0; i<numberOfRuns; i++){
            start=System.currentTimeMillis();
            KeyPair = function.call();
            finish=System.currentTimeMillis();
            System.out.println(i+") "+functionName+":\t"+
                    (finish-start) + "ms" +
                    "\tPrivKeySize: " + KeyPair[0].bitLength() +
                    "\tPubKeySize: " + KeyPair[1].bitLength() +
                    "\tProductSize: " + (KeyPair[0].multiply(KeyPair[1]).bitLength()));
            trialTimesList.add(finish-start);
        }
        for(int i = 0; i < numberOfRuns; i++) trialTimes[i] = trialTimesList.get(i);
        System.out.println("Durschnittliche Laufzeit: " + Arrays.stream(trialTimes).average().getAsDouble() + "ms");
        System.out.println("-----------------------------------------------------");
    }
    
    private static BigInteger[] generateRSAWithBigInteger(int bitLength){
        SecureRandom random = new SecureRandom();
        BigInteger publicKey = BigInteger.TEN;
        BigInteger privateKey = BigInteger.TEN;
        BigInteger product =  BigInteger.ZERO;
        //verify that the resulting bit length is correct
        while(product.bitLength()!=bitLength){
            publicKey=BigInteger.TEN;
            privateKey=BigInteger.TEN;
            //find a public key
            while(!publicKey.isProbablePrime(50)){
                byte[] bytes = new byte[bitLength/16];
                random.nextBytes(bytes);
                publicKey = new BigInteger(1,bytes);
            }
            //find a private key
            while(!privateKey.isProbablePrime(50)){
                byte[] bytes = new byte[bitLength/16];
                random.nextBytes(bytes);
                privateKey = new BigInteger(1,bytes);
            }
            product = privateKey.multiply(publicKey);
        }
        return new BigInteger[]{privateKey,publicKey};
    }
}
